# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import re
from pytest import raises

from jenkinsflow.flow import serial, parallel, FailedChildJobException
from .cfg import ApiType
from .framework import api_select
from .framework.utils import lines_in, replace_host_port, result_msg, build_started_msg


def test_reporting_job_status(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j12', max_fails=0, expect_invocations=1, invocation_delay=1.0, exec_time=1.5, initial_buildno=7, expect_order=2, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=0.5/api.speedup) as ctrl1:
            ctrl1.invoke('j11')
            ctrl1.invoke('j12')

        sout, _ = capsys.readouterr()

        if api.api_type == ApiType.MOCK:
            repr_not_invoked = "job: 'jenkinsflow_test__reporting_job_status__j11' Status IDLE - latest build: "
            assert repr_not_invoked in sout, repr_not_invoked + "\n - NOT FOUND IN:\n" + sout
            assert lines_in(api_type, sout, "job: 'jenkinsflow_test__reporting_job_status__j12' Status IDLE - latest build: #7")
            assert lines_in(api_type, sout, "'jenkinsflow_test__reporting_job_status__j12' Status QUEUED - Why am I queued?")
            assert lines_in(api_type, sout, "'jenkinsflow_test__reporting_job_status__j12' Status RUNNING - build: #8")
            assert lines_in(api_type, sout, "'jenkinsflow_test__reporting_job_status__j12' Status IDLE - build: #8")
        else:
            # TODO: know if we cleaned jobs and check the 'repr_not_invoked' above
            assert "'jenkinsflow_test__reporting_job_status__j12' Status RUNNING - build: " in sout
            # assert "'jenkinsflow_test__reporting_job_status__j12' Status IDLE - build: " in sout


def test_reporting_invocation_serial(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j12', max_fails=0, expect_invocations=1, invocation_delay=1.0, exec_time=1.5, initial_buildno=7, expect_order=2, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=0.5/api.speedup) as ctrl1:
            ctrl1.invoke('j11')
            ctrl1.invoke('j12')

        sout, _ = capsys.readouterr()
        empty_re = re.compile("^$")
        assert lines_in(
            api_type, sout,
            "^Defined Invocation http://x.x/job/jenkinsflow_test__reporting_invocation_serial__j11",
            "^Defined Invocation http://x.x/job/jenkinsflow_test__reporting_invocation_serial__j12",
            empty_re,
            "--- Starting flow ---",
            empty_re,
            "^Flow Invocation (1/1,1/1): ['jenkinsflow_test__reporting_invocation_serial__j11', 'jenkinsflow_test__reporting_invocation_serial__j12']",
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__reporting_invocation_serial__j11",
            build_started_msg(api, "jenkinsflow_test__reporting_invocation_serial__j11", 1),
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__reporting_invocation_serial__j12",
            build_started_msg(api, "jenkinsflow_test__reporting_invocation_serial__j12", 8),
        )


def test_reporting_invocation_parallel(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j12', max_fails=0, expect_invocations=1, invocation_delay=1.0, exec_time=1.5, initial_buildno=7, expect_order=2)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=0.5/api.speedup) as ctrl1:
            ctrl1.invoke('j11')
            ctrl1.invoke('j12')

        sout, _ = capsys.readouterr()
        assert lines_in(
            api_type, sout,
            "^Flow Invocation (1/1,1/1): ('jenkinsflow_test__reporting_invocation_parallel__j11', 'jenkinsflow_test__reporting_invocation_parallel__j12')",
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__reporting_invocation_parallel__j11",
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__reporting_invocation_parallel__j12",
        )

        assert lines_in(
            api_type, sout,
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__reporting_invocation_parallel__j11",
            build_started_msg(api, "jenkinsflow_test__reporting_invocation_parallel__j11", 1),
        )

        assert lines_in(
            api_type, sout,
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__reporting_invocation_parallel__j12",
            build_started_msg(api, "jenkinsflow_test__reporting_invocation_parallel__j12", 8)
        )


def test_reporting_multiple_invocations(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=3, expect_order=None)
        api.job('j12', max_fails=0, expect_invocations=1, invocation_delay=1.0, exec_time=1.5, initial_buildno=7, expect_order=4, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=0.5/api.speedup) as ctrl1:
            ctrl1.invoke('j11')
            ctrl1.invoke('j11')
            ctrl1.invoke('j11')
            ctrl1.invoke('j12')

        sout, _ = capsys.readouterr()
        empty_re = re.compile("^$")
        assert lines_in(
            api_type, sout,
            "^Defined Invocation-1 http://x.x/job/jenkinsflow_test__reporting_multiple_invocations__j11",
            "^Defined Invocation-2 http://x.x/job/jenkinsflow_test__reporting_multiple_invocations__j11",
            "^Defined Invocation-3 http://x.x/job/jenkinsflow_test__reporting_multiple_invocations__j11",
            "^Defined Invocation http://x.x/job/jenkinsflow_test__reporting_multiple_invocations__j12",
            empty_re,
            "--- Starting flow ---",
            empty_re,
            "^Flow Invocation (1/1,1/1): ['jenkinsflow_test__reporting_multiple_invocations__j11', 'jenkinsflow_test__reporting_multiple_invocations__j11', 'jenkinsflow_test__reporting_multiple_invocations__j11', 'jenkinsflow_test__reporting_multiple_invocations__j12']",
            "^Job Invocation-1 (1/1,1/1): http://x.x/job/jenkinsflow_test__reporting_multiple_invocations__j11",
            build_started_msg(api, "jenkinsflow_test__reporting_multiple_invocations__j11", 1, invocation_number=1),
            "^Job Invocation (1/1,1/1): http://x.x/job/jenkinsflow_test__reporting_multiple_invocations__j12",
            build_started_msg(api, "jenkinsflow_test__reporting_multiple_invocations__j12", 8),
        )


def test_reporting_retry(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11_fail', max_fails=1, expect_invocations=2, expect_order=1)
        api.job('j12', max_fails=0, expect_invocations=1, expect_order=2, serial=True)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=3, serial=True)
        api.job('j22_fail', max_fails=2, expect_invocations=3, expect_order=3)
        api.job('j31_fail', max_fails=3, expect_invocations=4, expect_order=3)
        api.job('j32', max_fails=0, expect_invocations=1, expect_order=3, serial=True)
        api.job('j23', max_fails=0, expect_invocations=1, expect_order=3)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=4, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
            ctrl1.invoke('j11_fail')
            ctrl1.invoke('j12')

            with ctrl1.parallel(timeout=70, max_tries=3) as ctrl2:
                ctrl2.invoke('j21')
                ctrl2.invoke('j22_fail')
                with ctrl2.serial(timeout=70, max_tries=2) as ctrl3:
                    ctrl3.invoke('j31_fail')
                    ctrl3.invoke('j32')
                ctrl2.invoke('j23')

            ctrl1.invoke('j13')

        sout, _ = capsys.readouterr()
        outer_flow_repr = "['jenkinsflow_test__reporting_retry__j11_fail', 'jenkinsflow_test__reporting_retry__j12', " \
                          "('jenkinsflow_test__reporting_retry__j21', 'jenkinsflow_test__reporting_retry__j22_fail', " \
                          "['jenkinsflow_test__reporting_retry__j31_fail', 'jenkinsflow_test__reporting_retry__j32'], " \
                          "'jenkinsflow_test__reporting_retry__j23'), 'jenkinsflow_test__reporting_retry__j13']"
        assert lines_in(
            api_type, sout,
            "^Flow Invocation (1/2,1/2): " + outer_flow_repr,
            "^Job Invocation (1/2,1/2): http://x.x/job/jenkinsflow_test__reporting_retry__j11_fail",
            build_started_msg(api, "jenkinsflow_test__reporting_retry__j11_fail", 1),
            "^FAILURE: 'jenkinsflow_test__reporting_retry__j11_fail'",
            "^RETRY: job: 'jenkinsflow_test__reporting_retry__j11_fail' failed, retrying child jobs from beginning. Up to 1 more times in current flow",
            "^Job Invocation (2/2,2/2): http://x.x/job/jenkinsflow_test__reporting_retry__j11_fail",
            build_started_msg(api, "jenkinsflow_test__reporting_retry__j11_fail", 2),
            "^SUCCESS: 'jenkinsflow_test__reporting_retry__j11_fail'",
            "^Job Invocation (1/3,1/6): http://x.x/job/jenkinsflow_test__reporting_retry__j23",
            build_started_msg(api, "jenkinsflow_test__reporting_retry__j23", 1),
            "^SUCCESS: 'jenkinsflow_test__reporting_retry__j23'",
            "^Job Invocation (1/2,1/2): http://.x.x/job/jenkinsflow_test__reporting_retry__j13",
            build_started_msg(api, "jenkinsflow_test__reporting_retry__j13", 1),
            "^Flow SUCCESS " + outer_flow_repr
        )

        assert lines_in(
            api_type, sout,
            "^Job Invocation (1/2,1/12): http://x.x/job/jenkinsflow_test__reporting_retry__j31_fail",
            build_started_msg(api, "jenkinsflow_test__reporting_retry__j31_fail", 1),
            "^FAILURE: 'jenkinsflow_test__reporting_retry__j31_fail'",
            "^RETRY: job: 'jenkinsflow_test__reporting_retry__j31_fail' failed, retrying child jobs from beginning. Up to 1 more times in current flow",
            "^Job Invocation (2/2,2/12): http://.x.x/job/jenkinsflow_test__reporting_retry__j31_fail",
            build_started_msg(api, "jenkinsflow_test__reporting_retry__j31_fail", 2),
            "^FAILURE: 'jenkinsflow_test__reporting_retry__j31_fail'",
            "^RETRY: job: 'jenkinsflow_test__reporting_retry__j31_fail' failed, retrying child jobs from beginning. Up to 10 more times through outer flow",
            "^Job Invocation (1/2,3/12): http://x.x/job/jenkinsflow_test__reporting_retry__j31_fail",
            build_started_msg(api, "jenkinsflow_test__reporting_retry__j31_fail", 3),
            "^FAILURE: 'jenkinsflow_test__reporting_retry__j31_fail'",
            "^RETRY: job: 'jenkinsflow_test__reporting_retry__j31_fail' failed, retrying child jobs from beginning. Up to 1 more times in current flow",
            "^Job Invocation (2/2,4/12): http://x.x/job/jenkinsflow_test__reporting_retry__j31_fail",
            build_started_msg(api, "jenkinsflow_test__reporting_retry__j31_fail", 4),
            "^SUCCESS: 'jenkinsflow_test__reporting_retry__j31_fail'"
        )


def test_reporting_result_unchecked(api_type, capsys):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21_unchecked', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=50, initial_buildno=7, expect_order=None, unknown_result=True, serial=True)
        api.job('j22', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=1.5, initial_buildno=7, expect_order=2)
        api.job('j31', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=1.5, initial_buildno=7, expect_order=3)
        api.job('j32_unchecked_fail', max_fails=1, expect_invocations=1, invocation_delay=0, exec_time=1.5, initial_buildno=7, expect_order=3)
        api.job('j41', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=1.5, initial_buildno=7, expect_order=3)
        api.job('j42_unchecked', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=1.5, initial_buildno=7, expect_order=3, serial=True)
        api.job('j23', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=1.5, initial_buildno=7, expect_order=4)
        api.job('j12', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=5, initial_buildno=7, expect_order=5)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=0.5/api.speedup) as ctrl1:
            ctrl1.invoke('j11')
            with ctrl1.serial() as ctrl2:
                ctrl2.invoke_unchecked('j21_unchecked')
                ctrl2.invoke('j22')
                with ctrl2.parallel() as ctrl3:
                    ctrl3.invoke('j31')
                    ctrl3.invoke_unchecked('j32_unchecked_fail')
                    with ctrl3.serial() as ctrl4:
                        ctrl4.invoke('j41')
                        ctrl4.invoke_unchecked('j42_unchecked')
                ctrl2.invoke('j23')
            ctrl1.invoke('j12')

        sout, _ = capsys.readouterr()

        assert lines_in(
            api_type, sout,
            "^UNCHECKED FAILURE: " + result_msg(api, "jenkinsflow_test__reporting_result_unchecked__j32_unchecked_fail"),
            "^UNCHECKED SUCCESS: " + result_msg(api, "jenkinsflow_test__reporting_result_unchecked__j42_unchecked", 8),
        )

def test_reporting_defined_job_parameters(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, invocation_delay=1.0, exec_time=1.5, initial_buildno=7, expect_order=1, serial=True,
                params=(('s1', '', 'desc'), ('c1', 'what', 'desc'), ('i1', 1, 'integer'), ('b1', False, 'boolean')))

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=0.5/api.speedup) as ctrl1:
            ctrl1.invoke('j1', s1="hi", c1='why?', i1=2, b1=True)

        sout, _ = capsys.readouterr()
        assert lines_in(
            api_type, sout,
            "^Defined Invocation http://x.x/job/jenkinsflow_test__reporting_defined_job_parameters__j1 - parameters:",
            "    i1 = '2'",
        )

        assert "    s1 = 'hi'" in sout
        assert "    c1 = 'why?'" in sout
        assert "    b1 = 'true'" in sout


ordered_params_output = """
Defined Invocation http://x.x/job/jenkinsflow_test__reporting_ordered_job_parameters__j1 - parameters:
     s1 = 'hi'
     s2 = 'not-last'
     c1 = 'why?'
     i1 = '2'
     b1 = 'true'
     s4 = 'was last'
     aaa = '3'
     unknown1 = 'Hello'
     unknown2 = 'true'
     s3 = 'last'
"""

def test_reporting_ordered_job_parameters(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, invocation_delay=1.0, exec_time=1.5, initial_buildno=7, expect_order=1, serial=True,
                params=(('s1', '', 'desc'), ('c1', 'what', 'desc'), ('i1', 1, 'integer'), ('b1', False, 'boolean'), ('s2', 't', 'd'), ('s3', 't2', 'd2'),
                        ('unknown1', 'Hello', 'd'), ('aaa', 17, 'd'), ('unknown2', False, 'd')))

        order = ['s1', 's2', 'c1', 'i1', 'b1', 's4', '*', 's3']
        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=0.5/api.speedup, params_display_order=order) as ctrl1:
            ctrl1.invoke('j1', s1="hi", c1='why?', i1=2, b1=True, s2='not-last', s3='last', unknown1='Hello', aaa=3, unknown2=True, s4='was last')

        sout, _ = capsys.readouterr()
        assert replace_host_port(api_type, ordered_params_output.strip()) in replace_host_port(api_type, sout)


def test_reporting_defined_non_existing(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        # TODO
        with raises(FailedChildJobException):
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=0.5/api.speedup, allow_missing_jobs=True) as ctrl1:
                ctrl1.invoke('j1', a="b", c='d')

        sout, _ = capsys.readouterr()
        assert lines_in(
            api_type, sout,
            "Defined Invocation 'jenkinsflow_test__reporting_defined_non_existing__j1' - MISSING JOB",
            "    a = 'b'",
        )
        assert "    c = 'd'" in sout



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function

import os
from os.path import join as jp
import re

from jenkinsflow.flow import serial

from .framework import api_select, utils
from .framework.utils import flow_graph_dir
from .cfg import ApiType

here = os.path.abspath(os.path.dirname(__file__))


_timestamp_re = re.compile(r't": [0-9]+.[0-9]+')


with open(jp(here, "json_test_compact.json")) as _jf:
    _compact_json = _jf.read().strip()


with open(jp(here, "json_test_pretty.json")) as _jf:
    _pretty_json = _jf.read().strip()


def _assert_json(got_json, expected_json, api_type):
    got_json = utils.replace_host_port(api_type, got_json)
    got_json = _timestamp_re.sub(r't": 12345.123', got_json)

    if api_type == ApiType.SCRIPT:
        expected_json = utils.replace_host_port(api_type, expected_json)

    if got_json.strip() != expected_json:
        if not os.environ.get('JOB_NAME'):
            print("--- expected json ---")
            print(expected_json)
            print("--- got json ---")
            print(got_json)
        assert got_json.strip() == expected_json
            

def _flow(api, strip_prefix, json_dir):
    if not os.path.exists(json_dir):
        os.makedirs(json_dir)

    with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1, json_dir=json_dir, json_strip_top_level_prefix=strip_prefix) as ctrl1:
        ctrl1.invoke('j1')
        ctrl1.invoke('j2')

        with ctrl1.parallel(timeout=40, report_interval=3) as ctrl2:
            with ctrl2.serial(timeout=40, report_interval=3) as ctrl3a:
                ctrl3a.invoke('j3')
                ctrl3a.invoke('j6')
                ctrl3a.invoke_unchecked('j7_unchecked')

            with ctrl2.parallel(timeout=40, report_interval=3) as ctrl3b:
                ctrl3b.invoke('j4')
                ctrl3b.invoke('j5')
                ctrl3b.invoke_unchecked('j8_unchecked')

        ctrl1.invoke('j9')

    return ctrl1


def test_json_strip_prefix(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        flow_name = api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j2', max_fails=0, expect_invocations=1, expect_order=2)
        api.job('j3', max_fails=0, expect_invocations=1, expect_order=3)
        api.job('j4', max_fails=0, expect_invocations=1, expect_order=3)
        api.job('j5', max_fails=0, expect_invocations=1, expect_order=3)
        api.job('j6', max_fails=0, expect_invocations=1, expect_order=3)
        api.job('j7_unchecked', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=40, expect_order=None, unknown_result=True)
        api.job('j8_unchecked', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=40, expect_order=None, unknown_result=True)
        api.job('j9', max_fails=0, expect_invocations=1, expect_order=4, exec_time=5)

        ctrl1 = _flow(api, True, flow_graph_dir(flow_name))

        # Test pretty printing
        json_file = jp(flow_graph_dir(flow_name), "pretty.json")
        ctrl1.json(json_file, indent=4)
        with open(json_file) as jf:
            _assert_json(jf.read().strip(), _pretty_json, api.api_type)

        # Test default compact json
        with open(ctrl1.json_file) as jf:
            _assert_json(jf.read().strip(), _compact_json, api.api_type)

        # Test return json
        json = ctrl1.json(None)
        _assert_json(json, _compact_json, api.api_type)


def test_json_no_strip_prefix(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        flow_name = api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j2', max_fails=0, expect_invocations=1, expect_order=2)
        api.job('j3', max_fails=0, expect_invocations=1, expect_order=3)
        api.job('j4', max_fails=0, expect_invocations=1, expect_order=3)
        api.job('j5', max_fails=0, expect_invocations=1, expect_order=3)
        api.job('j6', max_fails=0, expect_invocations=1, expect_order=3)
        api.job('j7_unchecked', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=40, expect_order=None, unknown_result=True)
        api.job('j8_unchecked', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=40, expect_order=None, unknown_result=True)
        api.job('j9', max_fails=0, expect_invocations=1, expect_order=4, exec_time=5)

        ctrl1 = _flow(api, False, flow_graph_dir(flow_name))

        # Test pretty printing with no stripping of top level prefix
        json_file = jp(flow_graph_dir(flow_name), "verbose_pretty.json")
        ctrl1.json(json_file, indent=4)
        with open(json_file) as jf:
            got_json = jf.read().strip()
            expect_json = _pretty_json.replace('strip_prefix', 'no_strip_prefix').replace('name": "', 'name": "jenkinsflow_test__json_no_strip_prefix__')
            _assert_json(got_json, expect_json, api.api_type)


def test_json_unchecked_only_in_flows(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        flow_name = api.flow_job()
        api.job('j1_unchecked', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=40, expect_order=None, unknown_result=True)
        api.job('j2_unchecked', max_fails=0, expect_invocations=1, expect_order=None)
        api.job('j3_unchecked', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=40, expect_order=None, unknown_result=True)
        api.job('j4_unchecked', max_fails=0, expect_invocations=1, invocation_delay=0, exec_time=40, expect_order=None, unknown_result=True)
        api.job('j5_unchecked', max_fails=0, expect_invocations=1, expect_order=None)
        api.job('j6', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j7', max_fails=0, expect_invocations=1, expect_order=2, exec_time=5)

        json_dir = flow_graph_dir(flow_name)
        if not os.path.exists(json_dir):
            os.makedirs(json_dir)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1, json_dir=json_dir) as ctrl1:
            ctrl1.invoke_unchecked('j1_unchecked')

            with ctrl1.parallel(timeout=40, report_interval=3) as ctrl2:
                with ctrl2.serial(timeout=40, report_interval=3) as ctrl3a:
                    ctrl3a.invoke_unchecked('j2_unchecked')
                    ctrl3a.invoke_unchecked('j3_unchecked')

                with ctrl2.parallel(timeout=40, report_interval=3) as ctrl3b:
                    ctrl3b.invoke_unchecked('j4_unchecked')
                    ctrl3b.invoke_unchecked('j5_unchecked')

            ctrl1.invoke('j6')
            ctrl1.invoke('j7')

        # Test default compact json
        with open(ctrl1.json_file) as got_jf, open(jp(here, "json_test_unchecked_compact.json")) as expected_jf:
            _assert_json(got_jf.read().strip(), expected_jf.read().strip(), api.api_type)



# Copyright (c) 2017 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import pytest

from jenkinsflow.flow import serial, FailedChildJobException, FailedChildJobsException

from .framework import api_select
from .cfg import ApiType


def _job_name(api, short_name):
    return api.job_name_prefix + short_name, short_name


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_enable_disable_disable_enable(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        full_name, short_name = _job_name(api, 'j1')
        api.job(short_name, max_fails=0, expect_invocations=0, expect_order=None)

        api.poll()
        job = api.get_job(full_name)
        job.disable()
        job.enable()


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_enable_disable_invoke_disabled(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        full_name, short_name = _job_name(api, 'j1_disabled_invoke')
        api.job(short_name, max_fails=1, expect_invocations=0, expect_order=None)

        api.poll()
        job = api.get_job(full_name)
        job.disable()

        with pytest.raises(Exception):  # TODO improved error handlng when invoking disabled jobs
            with serial(api, timeout=50, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
                ctrl1.invoke('j1_disabled_invoke')

        job.enable()



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import serial, parallel
from .framework import api_select


def test_empty_flow_top_level_serial(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()

        with serial(api, timeout=70):
            pass


def test_empty_flow_top_level_parallel(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()

        with parallel(api, timeout=70):
            pass


def test_empty_flow_serial_parallel(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j2', max_fails=0, expect_invocations=1, expect_order=2, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('j1')
            with ctrl1.parallel():
                pass
            ctrl1.invoke('j2')


def test_empty_flow_parallel_serial(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j2', max_fails=0, expect_invocations=1, expect_order=1)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('j1')
            with ctrl1.serial():
                pass
            ctrl1.invoke('j2')


def test_empty_flow_mix(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j31', max_fails=0, expect_invocations=1, expect_order=2)
        api.job('j32', max_fails=0, expect_invocations=1, expect_order=3, serial=True)
        api.job('j12', max_fails=0, expect_invocations=1, expect_order=4, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('j11')

            with ctrl1.parallel():
                pass

            with ctrl1.serial():
                pass

            with ctrl1.parallel() as ctrl2:
                with ctrl2.serial() as ctrl3a:
                    ctrl3a.invoke('j31')
                    ctrl3a.invoke('j32')

                with ctrl2.parallel():
                    pass

            ctrl1.invoke('j12')



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function
import os, re

import pytest

from ..lines_in import lines_in


_text = """
abcd1
bcde2
cdef3
defg4
efgh5
"""


def _no_output(capsys):
    sout, serr = capsys.readouterr()
    if sout or serr:
        return False
    return True


def _serr(capsys):
    sout, serr = capsys.readouterr()
    assert not sout
    return serr


def test_consecutive(capsys):
    assert lines_in(
        _text, None,
        'abcd1',
        'bcde2',
        'cdef3',
        'defg4',
        'efgh5',
    )
    assert _no_output(capsys)


def test_in_order(capsys):
    assert lines_in(
        _text, None,
        'abcd1',
        'cdef3',
        'efgh5',
    )
    assert _no_output(capsys)


def test_in_order_partial_line(capsys):
    assert lines_in(
        _text, None,
        'abc',
        'cde2',
        'fgh',
    )
    assert _no_output(capsys)


def test_in_order_start_line(capsys):
    assert lines_in(
        _text, None,
        '^abcd1',
        '^cdef3',
        '^efgh5',
    )
    assert _no_output(capsys)


def test_in_order_start_line_partial(capsys):
    assert lines_in(
        _text, None,
        '^ab',
        '^cde',
        '^efgh',
    )
    assert _no_output(capsys)


_not_in_order_expect_msg = """
Matched 1 lines:

cdef3

The text:

abcd1

--- NOT FOUND OR OUT OF ORDER IN ---


abcd1
bcde2
cdef3
defg4
efgh5
"""


def test_not_in_order(capsys):
    assert not lines_in(
        _text, None,
        'cdef3',
        'abcd1',
        'efgh5',
    )
    assert _not_in_order_expect_msg in _serr(capsys)


_not_in_order_partial_line_expect_msg = """
Matched 1 lines:

bcde2

The text:

abc

--- NOT FOUND OR OUT OF ORDER IN ---


abcd1
bcde2
cdef3
defg4
efgh5
"""

def test_not_in_order_partial_line(capsys):
    assert not lines_in(
        _text, None,
        'cde2',
        'abc',
        'fgh',
    )
    assert _not_in_order_partial_line_expect_msg in _serr(capsys)


_in_order_not_start_line_expect_msg = """
The text:

bcd1

--- NOT FOUND, OUT OF ORDER or NOT AT START OF LINE in ---


abcd1
bcde2
cdef3
defg4
efgh5
"""

def test_in_order_not_start_line(capsys):
    assert not lines_in(
        _text, None,
        '^bcd1',
        '^cdef3',
        '^efgh5',
    )
    assert _in_order_not_start_line_expect_msg in _serr(capsys)


def test_in_order_regex(capsys):
    assert lines_in(
        _text, None,
        re.compile('^ab.d1$'),
        re.compile('.*def3$'),
        re.compile('e.*5'),
    )
    assert _no_output(capsys)


_not_in_order_regex_expect_msg = """
Matched 1 lines:

cdef3

The regex:

^ab.d1$

--- NOT MATCHED or OUT OF ORDER in ---


abcd1
bcde2
cdef3
defg4
efgh5
"""

def test_not_in_order_regex(capsys):
    assert not lines_in(
        _text, None,
        re.compile('.*def3$'),
        re.compile('^ab.d1$'),
        re.compile('e.*5'),
    )
    assert _not_in_order_regex_expect_msg in _serr(capsys)


_not_matched_regex_expect_msg = """
The regex:

.*def7$

--- NOT MATCHED or OUT OF ORDER in ---


abcd1
bcde2
cdef3
defg4
efgh5
"""

def test_not_matched_regex(capsys):
    assert not lines_in(
        _text, None,
        re.compile('.*def7$'),
    )
    assert _not_matched_regex_expect_msg in _serr(capsys)


# TODO improved messages for 'no-text'
_not_matched_regex_no_text_expect_msg = """
No lines matched. (Empty text)

The regex:

.*def7$

--- NOT MATCHED or OUT OF ORDER in ---
"""

def test_not_matched_regex_no_text(capsys):
    assert not lines_in(
        '', None,
        re.compile('.*def7$'),
    )
    assert _not_matched_regex_no_text_expect_msg in _serr(capsys)


def test_unordered_matched(capsys):
    assert lines_in(
        _text, None,
        'abcd1', (
            'bcde2',
            'cdef3',
            'defg4',
        ),
        'efgh5'
    )

    assert lines_in(
        _text, None,
        'abcd1', (
            'defg4',
            'bcde2',
            'cdef3',
        ),
        'efgh5'
    )

    assert lines_in(
        _text, None,
        'abcd1', (
            'defg4',
            'cdef3',
            'efgh5',
        ),
    )

    assert _no_output(capsys)


def test_unordered_not_matched(capsys):
    assert not lines_in(
        _text, None,
        'abcd1', (
            'bcde2',
            'cdef3',
            'defg17',
        ),
        'efgh5'
    )

    assert not lines_in(
        _text, None,
        '^bcd1', (
            'defg4',
            'bcde2',
            'cdef3',
        ),
        'efgh5'
    )

    assert not lines_in(
        _text, None,
        'abcd1', (
            re.compile('efg4'),
            'cdef3',
            'efgh5',
        ),
    )

    assert not lines_in(
        _text, None,
        (
            re.compile('efg4'),
            'cdef3',
            'efgh5',
        ),
    )


def test_unordered_multiple_not_matched(capsys):
    assert not lines_in(
        _text, None,
        'abcd1', (
            'bcde2',
            'defg4',
            'cdef3',
        ), (
            'bcde2',
            'cdef33',
            'defg4',
        ),
        'efgh5'
    )

    assert not lines_in(
        _text, None,
        'abcd1', (
            'bcde2',
            'defg4',
            'cdef3',
        ), (
            'bcde2',
            'cdef3',
            'defg44',
        ),
    )


_repeat_text = """
XX
XX
"""

def test_in_order_repeat(capsys):
    assert lines_in(
        _repeat_text, None,
        "XX",
        "XX",
    )
    assert _no_output(capsys)


def test_unordered_repeat(capsys):
    assert lines_in(
        _repeat_text, None, (
            "XX",
            "XX",
        ),
    )
    assert _no_output(capsys)


def test_in_order_repeat_exhausted(capsys):
    assert not lines_in(
        _repeat_text, None,
        "XX",
        "XX",
        "XX",
    )


def test_unordered_repeat_exhausted(capsys):
    assert not lines_in(
        _repeat_text, None, (
            "XX",
            "XX",
            "XX",
        ),
    )


def test_in_order_unordered_repeat_exhausted(capsys):
    assert not lines_in(
        _repeat_text, None, (
            "XX",
            "XX",
        ),
        "XX",
    )


_long_text = """
abcd1
bcde2
cdef3
defg4
efgh5
6efgh
7efgh
8efgh
"""

def test_multi_nested(capsys):
    assert lines_in(
        _long_text, None,
        "abcd1", (
            "cdef3",
            "bcde2", (
                "defg4",
                "efgh5", (
                    '8efgh',
                    '7efgh',
                ),
            ),
        ),
    )
    assert _no_output(capsys)


def test_multi_nested_not_found():
    assert not lines_in(
        _long_text, None,
        "abcd1", (
            "cdef3",
            "bcde2", (
                "defg17",
                "efgh5", (
                    '8efgh',
                    '7efgh',
                ),
            ),
        ),
    )


def test_func_regex(capsys):
    def repl(line):
        return line.replace('bc', 'yy').replace('h5', 'yy')

    assert lines_in(
        _text, repl,
        re.compile('^ayyd1$'),
        'bcde2',
        'yy',
    )
    assert _no_output(capsys)


_func_regex_not_found_expect_msg = """
Matched 2 lines ('mfunc' modified line in '()'):

abcd1 (ayyd1)
bcde2 (yyde2)

The text:

yyz (yyz)

--- NOT FOUND OR OUT OF ORDER IN (unmodified text, 'mfunc' not applied) ---


abcd1
bcde2
cdef3
defg4
efgh5
"""

def test_func_regex_not_found(capsys):
    def repl(line):
        return line.replace('bc', 'yy').replace('h5', 'yy')

    assert not lines_in(
        _text, repl,
        re.compile('^ayyd1$'),
        'bcde2',
        'yyz',
    )
    assert _func_regex_not_found_expect_msg in _serr(capsys)
